# -*- coding: utf-8 -*-
"""
For testing neuromaps.nulls.nulls functionality
"""

import pytest


@pytest.mark.xfail
def test_alexander_bloch():
    assert False


@pytest.mark.xfail
def test_vasa():
    assert False


@pytest.mark.xfail
def test_hungarian():
    assert False


@pytest.mark.xfail
def test_baum():
    assert False


@pytest.mark.xfail
def test_cornblath():
    assert False


@pytest.mark.xfail
def test__get_distmat():
    assert False


@pytest.mark.xfail
def test__make_surrogates():
    assert False


@pytest.mark.xfail
def test_burt2018():
    assert False


@pytest.mark.xfail
def test_burt2020():
    assert False


@pytest.mark.xfail
def test_moran():
    assert False
