# -*- coding: utf-8 -*-
"""
For testing neuromaps.parcellate functionality
"""

import pytest


@pytest.mark.xfail
def test__gifti_to_array():
    assert False


@pytest.mark.xfail
def test__array_to_gifti():
    assert False


@pytest.mark.xfail
def test_Parcellater():
    assert False
