# -*- coding: utf-8 -*-
"""
For testing neuromaps.civet functionality
"""

import pytest


@pytest.mark.xfail
def test_read_civet_surf():
    assert False


@pytest.mark.xfail
def test_read_surfmap():
    assert False


@pytest.mark.xfail
def test_resample_surface_map():
    assert False
