# -*- coding: utf-8 -*-
"""
For testing neuromaps.caret functionality
"""

import pytest


@pytest.mark.xfail
def test_read_surface_shape():
    assert False


@pytest.mark.xfail
def test_read_coords():
    assert False


@pytest.mark.xfail
def test_read_topo():
    assert False


@pytest.mark.xfail
def test_read_deform_map():
    assert False


@pytest.mark.xfail
def test_apply_deform_map():
    assert False
